/// <reference types="node" />

import http from 'http';
import type { IncomingMessage } from 'http';
import type { ServerResponse } from 'http';

/**
 * Any compatible Buffer instance.
 * This is a minimal stand-alone definition of a Buffer instance. The actual type is that exported by node's typings.
 */
declare interface Buffer_2 extends Uint8Array {
}

/** Wire format reader using node buffers. */
declare class BufferReader extends Reader {

    /**
     * Constructs a new buffer reader instance.
     * @param buffer Buffer to read from
     */
    constructor(buffer: Buffer_2);

    /**
     * Reads a sequence of bytes preceeded by its length as a varint.
     * @returns Value read
     */
    public bytes(): Buffer_2;
}

/** Wire format writer using node buffers. */
declare class BufferWriter extends Writer {

    /** Constructs a new buffer writer instance. */
    constructor();

    /**
     * Allocates a buffer of the specified size.
     * @param size Buffer size
     * @returns Buffer
     */
    public static alloc(size: number): Buffer_2;

    /**
     * Finishes the write operation.
     * @returns Finished buffer
     */
    public finish(): Buffer_2;
}

export declare class Channel implements IChannel {
    private agent;
    private callback?;
    constructor(url: string, options?: IOptions);
    onCall(callback: (req: any, res?: any, err?: any) => void): void;
    reset(url: string, options?: IOptions): void;
    rpcUnaryUnary(service: string, method: string, requestSerializer: ISerializer, responseDeserializeBinary: IDeserializer, request: any): Promise<any>;
    rpcUnaryStream(service: string, method: string, requestSerializer: ISerializer, responseDeserializeBinary: IDeserializer, request: any): Promise<IStream<any>>;
    rpcStreamStream(service: string, method: string, requestSerializer: ISerializer, responseDeserializeBinary: IDeserializer, request: any): Promise<IStream<any>>;
    rpcStreamUnary(service: string, method: string, requestSerializer: ISerializer, responseDeserializeBinary: IDeserializer, request: any): Promise<any>;
}

export declare class Future<T = any> {
    promise: Promise<T>;
    result?: T;
    error: any;
    private _resovle?;
    private _reject?;
    private _resolved;
    constructor();
    resolve(result: T): void;
    reject(error: any): void;
    get resolved(): boolean;
}

export declare interface IChannel {
    rpcUnaryUnary: IRpc<any>;
    rpcUnaryStream: IRpc<IStream<any>>;
    rpcStreamStream: IRpc<IStream<any>>;
    rpcStreamUnary: IRpc<any>;
}

declare type IDeserializer = (bytes: Uint8Array) => any;

declare interface IOptions {
    contentType?: "grpc" | "grpc-web" | "grpc-web+proto" | "json";
    authorizationToken?: string;
    fetch?: typeof fetch;
}

declare type IRpc<T> = (service: string, method: string, requestSerializer: ISerializer, responseDeserialize: IDeserializer, request: any) => Promise<T>;

declare type ISerializer = (message: any) => Uint8Array;

export declare type IServiceFactory<T = any> = (request: IncomingMessage, response?: ServerResponse) => T | Promise<T>;

declare type IServiceMeta = {
    [key: string]: [IDeserializer, ISerializer, boolean, boolean];
};

export declare interface IStream<T = any> {
    abort(reason: any): void;
    read(onMessage: (o: T) => void): Promise<void>;
    readToItorator(): AsyncIterableIterator<T>;
}

/**
 * Any compatible Long instance.
 * This is a minimal stand-alone definition of a Long instance. The actual type is that exported by long.js.
 */
declare interface Long {

    /** Low bits */
    low: number;

    /** High bits */
    high: number;

    /** Whether unsigned or not */
    unsigned: boolean;
}

/** Wire format reader using `Uint8Array` if available, otherwise `Array`. */
export declare class Reader {

    /**
     * Constructs a new reader instance using the specified buffer.
     * @param buffer Buffer to read from
     */
    constructor(buffer: Uint8Array);

    /** Read buffer. */
    public buf: Uint8Array;

    /** Read buffer position. */
    public pos: number;

    /** Read buffer length. */
    public len: number;

    /**
     * Creates a new reader using the specified buffer.
     * @param buffer Buffer to read from
     * @returns A {@link BufferReader} if `buffer` is a Buffer, otherwise a {@link Reader}
     * @throws {Error} If `buffer` is not a valid buffer
     */
    public static create(buffer: (Uint8Array|Buffer_2)): (Reader|BufferReader);

    /**
     * Reads a varint as an unsigned 32 bit value.
     * @returns Value read
     */
    public uint32(): number;

    /**
     * Reads a varint as a signed 32 bit value.
     * @returns Value read
     */
    public int32(): number;

    /**
     * Reads a zig-zag encoded varint as a signed 32 bit value.
     * @returns Value read
     */
    public sint32(): number;

    /**
     * Reads a varint as a signed 64 bit value.
     * @returns Value read
     */
    public int64(): Long;

    /**
     * Reads a varint as an unsigned 64 bit value.
     * @returns Value read
     */
    public uint64(): Long;

    /**
     * Reads a zig-zag encoded varint as a signed 64 bit value.
     * @returns Value read
     */
    public sint64(): Long;

    /**
     * Reads a varint as a boolean.
     * @returns Value read
     */
    public bool(): boolean;

    /**
     * Reads fixed 32 bits as an unsigned 32 bit integer.
     * @returns Value read
     */
    public fixed32(): number;

    /**
     * Reads fixed 32 bits as a signed 32 bit integer.
     * @returns Value read
     */
    public sfixed32(): number;

    /**
     * Reads fixed 64 bits.
     * @returns Value read
     */
    public fixed64(): Long;

    /**
     * Reads zig-zag encoded fixed 64 bits.
     * @returns Value read
     */
    public sfixed64(): Long;

    /**
     * Reads a float (32 bit) as a number.
     * @returns Value read
     */
    public float(): number;

    /**
     * Reads a double (64 bit float) as a number.
     * @returns Value read
     */
    public double(): number;

    /**
     * Reads a sequence of bytes preceeded by its length as a varint.
     * @returns Value read
     */
    public bytes(): Uint8Array;

    /**
     * Reads a string preceeded by its byte length as a varint.
     * @returns Value read
     */
    public string(): string;

    /**
     * Skips the specified number of bytes if specified, otherwise skips a varint.
     * @param [length] Length if known, otherwise a varint is assumed
     * @returns `this`
     */
    public skip(length?: number): Reader;

    /**
     * Skips the next element of the specified wire type.
     * @param wireType Wire type received
     * @returns `this`
     */
    public skipType(wireType: number): Reader;
}

export declare class Server {
    private services;
    addService(service: string, rpcs: IServiceMeta, factory: IServiceFactory): void;
    middlewares: (req: http.IncomingMessage, res: http.ServerResponse) => void;
    listenHttpServerRequest(server: http.Server, path: string, fallback?: (request: http.IncomingMessage, response: http.ServerResponse) => void): void;
    listenHttpServerUpgrade(server: http.Server, path: string, fallback?: (request: http.IncomingMessage, socket: any, upgradeHead: any) => void): void;
}

export declare class Stream<T = any> implements IStream {
    private _onabort?;
    private _readers;
    private _closed;
    private _cache;
    private _error;
    constructor(_onabort?: ((reason: any) => void) | undefined);
    write(o: T): void;
    error(e: any): void;
    end(): void;
    get closed(): boolean;
    abort(reason: any): void;
    read(onMessage: (o: T) => void): Promise<void>;
    readToItorator(): AsyncIterableIterator<T>;
    writeFromIterator(it: () => AsyncGenerator<T>): Promise<void>;
}

/** Wire format writer using `Uint8Array` if available, otherwise `Array`. */
export declare class Writer {

    /** Constructs a new writer instance. */
    constructor();

    /** Current length. */
    public len: number;

    /** Operations head. */
    public head: object;

    /** Operations tail */
    public tail: object;

    /** Linked forked states. */
    public states: (object|null);

    /**
     * Creates a new writer.
     * @returns A {@link BufferWriter} when Buffers are supported, otherwise a {@link Writer}
     */
    public static create(): (BufferWriter|Writer);

    /**
     * Allocates a buffer of the specified size.
     * @param size Buffer size
     * @returns Buffer
     */
    public static alloc(size: number): Uint8Array;

    /**
     * Writes an unsigned 32 bit value as a varint.
     * @param value Value to write
     * @returns `this`
     */
    public uint32(value: number): Writer;

    /**
     * Writes a signed 32 bit value as a varint.
     * @param value Value to write
     * @returns `this`
     */
    public int32(value: number): Writer;

    /**
     * Writes a 32 bit value as a varint, zig-zag encoded.
     * @param value Value to write
     * @returns `this`
     */
    public sint32(value: number): Writer;

    /**
     * Writes an unsigned 64 bit value as a varint.
     * @param value Value to write
     * @returns `this`
     * @throws {TypeError} If `value` is a string and no long library is present.
     */
    public uint64(value: (Long|number|string)): Writer;

    /**
     * Writes a signed 64 bit value as a varint.
     * @param value Value to write
     * @returns `this`
     * @throws {TypeError} If `value` is a string and no long library is present.
     */
    public int64(value: (Long|number|string)): Writer;

    /**
     * Writes a signed 64 bit value as a varint, zig-zag encoded.
     * @param value Value to write
     * @returns `this`
     * @throws {TypeError} If `value` is a string and no long library is present.
     */
    public sint64(value: (Long|number|string)): Writer;

    /**
     * Writes a boolish value as a varint.
     * @param value Value to write
     * @returns `this`
     */
    public bool(value: boolean): Writer;

    /**
     * Writes an unsigned 32 bit value as fixed 32 bits.
     * @param value Value to write
     * @returns `this`
     */
    public fixed32(value: number): Writer;

    /**
     * Writes a signed 32 bit value as fixed 32 bits.
     * @param value Value to write
     * @returns `this`
     */
    public sfixed32(value: number): Writer;

    /**
     * Writes an unsigned 64 bit value as fixed 64 bits.
     * @param value Value to write
     * @returns `this`
     * @throws {TypeError} If `value` is a string and no long library is present.
     */
    public fixed64(value: (Long|number|string)): Writer;

    /**
     * Writes a signed 64 bit value as fixed 64 bits.
     * @param value Value to write
     * @returns `this`
     * @throws {TypeError} If `value` is a string and no long library is present.
     */
    public sfixed64(value: (Long|number|string)): Writer;

    /**
     * Writes a float (32 bit).
     * @param value Value to write
     * @returns `this`
     */
    public float(value: number): Writer;

    /**
     * Writes a double (64 bit float).
     * @param value Value to write
     * @returns `this`
     */
    public double(value: number): Writer;

    /**
     * Writes a sequence of bytes.
     * @param value Buffer or base64 encoded string to write
     * @returns `this`
     */
    public bytes(value: (Uint8Array|string)): Writer;

    /**
     * Writes a string.
     * @param value Value to write
     * @returns `this`
     */
    public string(value: string): Writer;

    /**
     * Forks this writer's state by pushing it to a stack.
     * Calling {@link Writer#reset|reset} or {@link Writer#ldelim|ldelim} resets the writer to the previous state.
     * @returns `this`
     */
    public fork(): Writer;

    /**
     * Resets this instance to the last state.
     * @returns `this`
     */
    public reset(): Writer;

    /**
     * Resets to the last state and appends the fork state's current write length as a varint followed by its operations.
     * @returns `this`
     */
    public ldelim(): Writer;

    /**
     * Finishes the write operation.
     * @returns Finished buffer
     */
    public finish(): Uint8Array;
}

export { }
